/*
   BAREOS® - Backup Archiving REcovery Open Sourced

   Copyright (C) 2023-2023 Bareos GmbH & Co. KG

   This program is Free Software; you can redistribute it and/or
   modify it under the terms of version three of the GNU Affero General Public
   License as published by the Free Software Foundation and included
   in the file LICENSE.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA.
*/

#ifndef BAREOS_STORED_BACKENDS_DEDUP_UTIL_H_
#define BAREOS_STORED_BACKENDS_DEDUP_UTIL_H_

#include <fcntl.h>
#include <unistd.h>
#include <memory>
#include <utility>
#include <cstring>
#include <optional>
#include <sys/mman.h>
#include "lib/network_order.h"

namespace no = network_order;
namespace dedup::util {
static_assert(((ssize_t)(off_t)-1) < 0,
              "The code assumes that the error condition is negative when cast "
              "to ssize_t");
class raii_fd {
 public:
  raii_fd() = default;
  raii_fd(const char* path_, int flags_, int mode_)
      : path{path_}
      , flags{flags_}
      , mode{mode_}
      , fd{::open(path.c_str(), flags, mode)}
      , error{fd < 0}
  {
  }

  raii_fd(int dird, const char* path_, int flags_, int mode_)
      : path{path_}
      , flags{flags_}
      , mode{mode_}
      , fd{::openat(dird, path.c_str(), flags, mode)}
      , error{fd < 0}
  {
  }

  raii_fd(raii_fd&& move_from) : raii_fd{} { *this = std::move(move_from); }

  raii_fd& operator=(raii_fd&& move_from)
  {
    std::swap(fd, move_from.fd);
    std::swap(flags, move_from.flags);
    std::swap(mode, move_from.mode);
    std::swap(error, move_from.error);
    std::swap(path, move_from.path);
    return *this;
  }

  bool is_ok() const { return !(fd < 0 || error); }

  int get() const { return fd; }

  bool flush() { return ::fsync(fd) == 0; }

  bool resize(std::size_t new_size)
  {
    // this looks weird but is done for the following reason:
    // if new_size > current_size, then ftruncate would not actually allocate
    // the memory but instead make the file sparse!  This will reduce write
    // performance.  So in these cases we want to use fallocate.
    // if new_size < current_size, we want to use ftruncate to make the file
    // smaller.
    // Since fallocate does not do anything if new_size <= current_size,
    // and ftruncate does not do anything if current_size >= new_size,
    // we can accomplish what this function is trying to do by
    // first calling fallocate and then calling ftruncate.
    // we need to take into account that fallocate will error
    // if new_size is 0!

    // some filesystems do not support posix_fallocate (ex. zfs on FreeBsd),
    // so we ignore the error message here. so we ignore the error return here.
    // the fs should just take it as a hint that it would be nice if the space
    // actually existed if we need it.
    if (int error = 0;
        new_size != 0 && (error = posix_fallocate(fd, 0, new_size)) != 0) {
      Dmsg0(250, "posix_fallocate returned error: %d ERR=%s\n", error,
            strerror(error));
      // return false;
    }
    return ::ftruncate(fd, new_size) == 0;
  }

  bool write_at(std::size_t offset, const void* data, std::size_t size)
  {
    ssize_t res = ::pwrite(fd, data, size, offset);

    if (res < 0) {
      error = true;
      return false;
    }

    if (static_cast<std::size_t>(res) != size) { return false; }

    return true;
  }

  bool read_at(std::size_t offset, void* data, std::size_t size) const
  {
    ssize_t res = ::pread(fd, data, size, offset);

    if (res < 0) {
      error = true;
      return false;
    }

    if (static_cast<std::size_t>(res) != size) { return false; }

    return true;
  }

  std::optional<std::size_t> size()
  {
    ssize_t size = ::lseek(fd, 0, SEEK_END);
    if (size < 0) {
      error = true;
      return std::nullopt;
    }
    return static_cast<std::size_t>(size);
  }

  ~raii_fd()
  {
    if (fd >= 0) { close(fd); }
  }

  const char* relative_path() const { return path.c_str(); }

 private:
  std::string path{};
  int flags{};
  int mode{};
  int fd{-1};
  mutable bool error{true};
};

struct write_buffer {
  char* current;
  char* end;

  write_buffer(char* data, std::size_t size)
      : current{static_cast<char*>(data)}, end{current + size}
  {
  }

  bool write(std::size_t size, const char* data)
  {
    if (current + size > end) { return false; }

    current = std::copy(data, data + size, current);
    return true;
  }

  char* reserve(std::size_t size)
  {
    if (current + size > end) { return nullptr; }

    return std::exchange(current, current + size);
  }

  template <typename F, typename = std::enable_if_t<no::is_serializable_v<F>>>
  inline bool write(const F& val)
  {
    return write(sizeof(F), reinterpret_cast<const char*>(&val));
  }
};

template <typename T> class file_based_array {
  using needs_to_be_trivially_copyable
      = std::enable_if_t<std::is_trivially_copyable_v<T>, bool>;

 public:
  file_based_array() = default;
  file_based_array(raii_fd file, std::size_t used);
  file_based_array(file_based_array&& other) : file_based_array{}
  {
    *this = std::move(other);
  }
  ~file_based_array();
  file_based_array& operator=(file_based_array&& other);

  std::optional<std::size_t> reserve(std::size_t count);
  bool write_at(std::size_t start, const T* arr, std::size_t count);
  std::optional<std::size_t> push_back(const T& val)
  {
    return push_back(&val, 1);
  }
  std::optional<std::size_t> push_back(const T* arr, std::size_t count);
  bool write_at(std::size_t start, const T& val)
  {
    return write_at(start, &val, 1);
  }

  bool read_at(std::size_t start, T* arr, std::size_t count = 1) const;

  bool flush()
  {
    if (error) { return false; }
    // if we used a cache we would write it out here
    msync(memory, cap * elem_size, MS_SYNC);
    return file.flush();
  }

  inline void clear() { used = 0; }

  inline std::size_t size() const { return used; }
  inline std::size_t capacity() const { return cap; }

  inline bool is_ok() const { return !error && file.is_ok(); }

  static constexpr std::size_t elem_size = sizeof(T);

  const raii_fd& backing_file() const { return file; }

 private:
  std::size_t used{0};
  std::size_t cap;
  raii_fd file{};
  T* memory{nullptr};
  bool error{true};

  std::optional<std::size_t> reserve_at(std::size_t at, std::size_t count);
};

template <typename T>
std::optional<std::size_t> file_based_array<T>::reserve(std::size_t count)
{
  return reserve_at(used, count);
}

template <typename T>
std::optional<std::size_t> file_based_array<T>::reserve_at(std::size_t at,
                                                           std::size_t count)
{
  if (error) { return std::nullopt; }

  if (at + count < at) {
    return std::nullopt;  // make sure nothing weird is going on
  }

  if (at > used) {
    // since this is an internal function
    // this should never happen.  So if it does set the error flag
    error = true;
    return std::nullopt;
  }

  if (at + count > cap) { return std::nullopt; }

  used = std::max(used, at + count);
  return at;
}

template <typename T>
std::optional<std::size_t> file_based_array<T>::push_back(const T* arr,
                                                          std::size_t count)
{
  std::optional start = reserve_at(used, count);

  if (!start) { return std::nullopt; }

  if (!write_at(*start, arr, count)) { return std::nullopt; }

  return start;
}

template <typename T>
bool file_based_array<T>::write_at(std::size_t start,
                                   const T* arr,
                                   std::size_t count)
{
  if (error) { return false; }

  if (start > used) { return false; }

  std::memcpy(memory + start, arr, count * elem_size);

  return true;
}

template <typename T>
bool file_based_array<T>::read_at(std::size_t start,
                                  T* arr,
                                  std::size_t count) const
{
  if (error) { return false; }

  if (start + count > used) { return false; }

  std::copy_n(memory + start, count, arr);

  return true;
}

template <typename T>
file_based_array<T>::file_based_array(raii_fd file_, std::size_t used_)
    : used{used_}, cap{0}, file{std::move(file_)}, error{!file.is_ok()}
{
  if (error) { return; }

  // let us compute the capacity

  std::optional size = file.size();

  if (!size) {
    error = true;
    return;
  }

  cap = *size / elem_size;

  if (used > cap) {
    error = true;
    return;
  }

  // todo: use the permissions from file here somehow
  void* ptr = mmap(nullptr, cap * elem_size, PROT_READ | PROT_WRITE, MAP_SHARED,
                   file.get(), 0);

  if (ptr == MAP_FAILED) {
    error = true;
    return;
  }

  memory = static_cast<T*>(ptr);
}

template <typename T> file_based_array<T>::~file_based_array()
{
  if (memory) {
    flush();
    munmap(memory, cap * elem_size);
  }
}

template <typename T>
file_based_array<T>& file_based_array<T>::operator=(file_based_array<T>&& other)
{
  std::swap(used, other.used);
  std::swap(cap, other.cap);
  std::swap(file, other.file);
  std::swap(error, other.error);
  std::swap(memory, other.memory);
  return *this;
}

template <typename T> class file_based_vector {
  using needs_to_be_trivially_copyable
      = std::enable_if_t<std::is_trivially_copyable_v<T>, bool>;

 public:
  file_based_vector() = default;
  file_based_vector(raii_fd file,
                    std::size_t used,
                    std::size_t capacity_chunk_size);
  file_based_vector(file_based_vector&& other) : file_based_vector{}
  {
    *this = std::move(other);
  }
  file_based_vector& operator=(file_based_vector&& other);

  std::optional<std::size_t> reserve(std::size_t count);
  std::optional<std::size_t> push_back(const T* arr, std::size_t count);
  inline std::optional<std::size_t> push_back(const T& val)
  {
    return push_back(&val, 1);
  }

  inline bool write_at(std::size_t start, const T& val)
  {
    return write_at(start, &val, 1);
  }

  bool write_at(std::size_t start, const T* arr, std::size_t count);
  bool read_at(std::size_t start, T* arr, std::size_t count) const;

  bool flush()
  {
    if (error) { return false; }
    // if we used a cache we would write it out here
    return file.flush();
  }

  bool shrink_to_fit();

  inline void clear() { used = 0; }

  inline std::size_t size() const { return used; }

  inline bool is_ok() const { return !error && file.is_ok(); }

  const raii_fd& backing_file() const { return file; }

 private:
  std::size_t used{0};
  std::size_t capacity;
  std::size_t capacity_chunk_size{1};
  raii_fd file;
  mutable bool error{true};
  static constexpr std::size_t elem_size = sizeof(T);
};

template <typename T>
std::optional<std::size_t> file_based_vector<T>::reserve(std::size_t count)
{
  if (error) { return std::nullopt; }

  std::size_t new_used = used + count;
  if (new_used < used) {
    return std::nullopt;  // make sure nothing weird is going on
  }

  if (new_used > capacity) {
    std::size_t delta = new_used - capacity;

    // compute first mutilple of capacity_chunk_size that is greater than delta
    std::size_t num_new_items
        = ((delta + capacity_chunk_size - 1) / capacity_chunk_size)
          * capacity_chunk_size;
    ASSERT(num_new_items + capacity >= new_used);

    std::size_t new_cap = capacity + num_new_items;
    if (new_cap < new_used) {
      // something weird is going on :S
      return std::nullopt;
    }

    if (!file.resize(new_cap * elem_size)) {
      error = true;
      return std::nullopt;
    }

    capacity = new_cap;
  }

  return std::exchange(used, new_used);
}

template <typename T>
std::optional<std::size_t> file_based_vector<T>::push_back(const T* arr,
                                                           std::size_t count)
{
  std::optional start = reserve(count);

  if (!start) { return false; }

  // write_at always returns to the current iter
  // if we set iter to the new desired position
  // we prevent the double seek we would need to do otherwise
  if (!write_at(*start, arr, count)) { return std::nullopt; }

  return start;
}

template <typename T>
bool file_based_vector<T>::write_at(std::size_t start,
                                    const T* arr,
                                    std::size_t count)
{
  if (error) { return false; }

  if (start > used) { return false; }

  if (start + count > used) { return false; }

  if (!file.write_at(elem_size * start, arr, count * elem_size)) {
    error = true;
    return false;
  }

  return true;
}

template <typename T>
bool file_based_vector<T>::read_at(std::size_t start,
                                   T* arr,
                                   std::size_t count) const
{
  if (error) { return false; }

  if (start + count > used) { return false; }

  if (!file.read_at(elem_size * start, arr, count * elem_size)) {
    error = true;
    return false;
  }

  return true;
}

template <typename T>
file_based_vector<T>::file_based_vector(raii_fd file_,
                                        std::size_t used_,
                                        std::size_t capacity_chunk_size_)
    : used{used_}
    , capacity{0}
    , capacity_chunk_size(capacity_chunk_size_)
    , file{std::move(file_)}
    , error{!file.is_ok()}
{
  if (error) { return; }

  // let us compute the capacity

  std::optional size = file.size();

  if (!size) {
    error = true;
    return;
  }

  capacity = *size / elem_size;

  if (used > capacity) {
    error = true;
    return;
  }
}

template <typename T> bool file_based_vector<T>::shrink_to_fit()
{
  capacity = used;
  return file.resize(capacity * elem_size);
}

template <typename T>
file_based_vector<T>& file_based_vector<T>::operator=(
    file_based_vector<T>&& other)
{
  std::swap(used, other.used);
  std::swap(capacity, other.capacity);
  std::swap(capacity_chunk_size, other.capacity_chunk_size);
  std::swap(file, other.file);
  std::swap(error, other.error);
  return *this;
}

}; /* namespace dedup::util */

#endif  // BAREOS_STORED_BACKENDS_DEDUP_UTIL_H_
